// $Id: HandleSupport.cpp,v 1.10 2002/03/04 23:59:39 cthuang Exp $
#include <sstream>
#include "CmdNameType.h"

InternalRep::InternalRep (
    Tcl_Interp *interp,
    Tcl_ObjCmdProc *pCmdProc):
        m_interp(interp),
        m_handleCount(0)
{
    std::string handleName(name());

    m_command = Tcl_CreateObjCommand(
        m_interp,
        const_cast<char *>(handleName.c_str()),
        pCmdProc,
        this,
        0);

    m_pNameEntry = HandleNameToRepMap::instance(m_interp)->insert(
        handleName.c_str(), this);
}

InternalRep::~InternalRep ()
{
    HandleNameToRepMap::instance(m_interp)->erase(m_pNameEntry);
    Tcl_DeleteCommandFromToken(m_interp, m_command);
}

std::string
InternalRep::name () const
{
    std::ostringstream oss;
    oss << "::counted::command0x" << std::hex << this;
    return oss.str();
}

void
InternalRep::incrHandleCount ()
{
    ++m_handleCount;
}

long
InternalRep::decrHandleCount ()
{
    if (--m_handleCount == 0) {
        delete this;
        return 0;
    }
    return m_handleCount;
}

// This maps Tcl_Obj pointers to an internal representation.

class ObjToRepMap
{
    struct ThreadSpecificData
    {
        Tcl_HashTable m_hashTable;
        bool m_initialized;
    };

    static Tcl_ThreadDataKey m_threadDataKey;

    static void exitProc(ClientData);
    static Tcl_HashTable *hashTable();

public:
    static void insert(Tcl_Obj *pObj, InternalRep *pRep);
    static InternalRep *find(Tcl_Obj *pObj);
    static void erase(Tcl_Obj *pObj);
};

Tcl_ThreadDataKey ObjToRepMap::m_threadDataKey;

void
ObjToRepMap::exitProc (ClientData)
{
    Tcl_DeleteHashTable(hashTable());
}

Tcl_HashTable *
ObjToRepMap::hashTable ()
{
    ThreadSpecificData *tsdPtr = static_cast<ThreadSpecificData *>(
        Tcl_GetThreadData(&m_threadDataKey, sizeof(ThreadSpecificData)));
    if (!tsdPtr->m_initialized) {
        tsdPtr->m_initialized = true;
        Tcl_InitHashTable(&tsdPtr->m_hashTable, TCL_ONE_WORD_KEYS);
        Tcl_CreateThreadExitHandler(exitProc, 0);
    }
    return &tsdPtr->m_hashTable;
}

void
ObjToRepMap::insert (Tcl_Obj *pObj, InternalRep *pRep)
{
    int isNew;
    Tcl_HashEntry *pEntry = Tcl_CreateHashEntry(
        hashTable(), reinterpret_cast<char *>(pObj), &isNew);
    Tcl_SetHashValue(pEntry, pRep);
}

InternalRep *
ObjToRepMap::find (Tcl_Obj *pObj)
{
    Tcl_HashEntry *pEntry = Tcl_FindHashEntry(
        hashTable(), reinterpret_cast<char *>(pObj));
    if (pEntry == 0) {
        return 0;
    }
    return static_cast<InternalRep *>(Tcl_GetHashValue(pEntry));
}

void
ObjToRepMap::erase (Tcl_Obj *pObj)
{
    Tcl_HashEntry *pEntry = Tcl_FindHashEntry(
        hashTable(), reinterpret_cast<char *>(pObj));
    if (pEntry != 0) {
        Tcl_DeleteHashEntry(pEntry);
    }
}


Tcl_ObjType *CmdNameType::ms_pCmdNameType;
Tcl_ObjType CmdNameType::ms_oldCmdNameType;

Singleton<CmdNameType> CmdNameType::ms_singleton;

CmdNameType &
CmdNameType::instance ()
{
    return ms_singleton.instance();
}

CmdNameType::CmdNameType ()
{
    // Hijack Tcl's cmdName type.
    ms_pCmdNameType = Tcl_GetObjType("cmdName");
    ms_oldCmdNameType = *ms_pCmdNameType;
    ms_pCmdNameType->freeIntRepProc = freeInternalRep;
    ms_pCmdNameType->dupIntRepProc = dupInternalRep;
    ms_pCmdNameType->updateStringProc = updateString;
    ms_pCmdNameType->setFromAnyProc = setFromAny;
}

CmdNameType::~CmdNameType ()
{
    // Restore original cmdName type.
    ms_pCmdNameType->freeIntRepProc = ms_oldCmdNameType.freeIntRepProc;
    ms_pCmdNameType->dupIntRepProc = ms_oldCmdNameType.dupIntRepProc;
    ms_pCmdNameType->updateStringProc = ms_oldCmdNameType.updateStringProc;
    ms_pCmdNameType->setFromAnyProc = ms_oldCmdNameType.setFromAnyProc;
}

void
CmdNameType::freeInternalRep (Tcl_Obj *pObj)
{
    if (pObj->refCount == 0) {
        InternalRep *pRep = ObjToRepMap::find(pObj);
        if (pRep != 0 && pRep->decrHandleCount() == 0) {
            ObjToRepMap::erase(pObj);
        }
    }

    ms_oldCmdNameType.freeIntRepProc(pObj);
}

void
CmdNameType::dupInternalRep (Tcl_Obj *pSrc, Tcl_Obj *pDup)
{
    // TODO: An object is duplicated if it is about to be modified.  We cannot
    // allow modifications on a handle.

    ms_oldCmdNameType.dupIntRepProc(pSrc, pDup);
}

// This should never be called because the string representation should already
// be valid.

void
CmdNameType::updateString (Tcl_Obj *pObj)
{
    ms_oldCmdNameType.updateStringProc(pObj);
}

int
CmdNameType::setFromAny (Tcl_Interp *interp, Tcl_Obj *pObj)
{
    // Check if the string represents an existing handle.
    InternalRep *pRep = 0;
    HandleNameToRepMap *pHandleNameToRepMap =
        HandleNameToRepMap::instance(interp);
    if (pHandleNameToRepMap != 0) {
        pRep = pHandleNameToRepMap->find(pObj);
        if (pRep != 0 && ObjToRepMap::find(pObj) == 0) {
            pRep->incrHandleCount();
        }
    }

    int result = ms_oldCmdNameType.setFromAnyProc(interp, pObj);

    if (result == TCL_OK && pRep != 0) {
        ObjToRepMap::insert(pObj, pRep);
    }
    return result;
}

Tcl_Obj *
CmdNameType::newObj (Tcl_Interp *interp, InternalRep *pRep)
{
    Tcl_Obj *pObj = Tcl_NewStringObj(
        const_cast<char *>(pRep->name().c_str()), -1);
    Tcl_ConvertToType(interp, pObj, ms_pCmdNameType);
    return pObj;
}


static char ASSOC_KEY[] = "countedHandles";

HandleNameToRepMap::HandleNameToRepMap (Tcl_Interp *interp)
{
    Tcl_InitHashTable(&m_handleToRepMap, TCL_STRING_KEYS);

    Tcl_SetAssocData(interp, ASSOC_KEY, deleteInterpProc, this);
    Tcl_CreateExitHandler(exitProc, this);
}

HandleNameToRepMap::~HandleNameToRepMap ()
{
    // Clean up any left over objects.
    clear();

    Tcl_DeleteHashTable(&m_handleToRepMap);
}

void
HandleNameToRepMap::clear ()
{
    Tcl_HashSearch search;
    Tcl_HashEntry *pEntry = Tcl_FirstHashEntry(&m_handleToRepMap, &search);
    while (pEntry != 0) {
        Tcl_HashEntry *pNext = Tcl_NextHashEntry(&search);
        delete static_cast<InternalRep *>(Tcl_GetHashValue(pEntry));
        pEntry = pNext;
    }
}

void
HandleNameToRepMap::deleteInterpProc (ClientData clientData, Tcl_Interp *)
{
    Tcl_DeleteExitHandler(exitProc, clientData);
    delete static_cast<HandleNameToRepMap *>(clientData);
}

void
HandleNameToRepMap::exitProc (ClientData clientData)
{
    delete static_cast<HandleNameToRepMap *>(clientData);
}

HandleNameToRepMap *
HandleNameToRepMap::instance (Tcl_Interp *interp)
{
    return static_cast<HandleNameToRepMap *>(
        Tcl_GetAssocData(interp, ASSOC_KEY, 0));
}

Tcl_HashEntry *
HandleNameToRepMap::insert (const char *handleStr, InternalRep *pRep)
{
    int isNew;
    Tcl_HashEntry *pEntry = Tcl_CreateHashEntry(
        &m_handleToRepMap,
        const_cast<char *>(handleStr),
        &isNew);
    Tcl_SetHashValue(pEntry, static_cast<ClientData>(pRep));
    return pEntry;
}

InternalRep *
HandleNameToRepMap::find (Tcl_Obj *pHandle) const
{
    char *key = Tcl_GetStringFromObj(pHandle, 0);
    Tcl_HashEntry *pEntry = Tcl_FindHashEntry(
        const_cast<Tcl_HashTable *>(&m_handleToRepMap),
        key);
    if (pEntry == 0) {
        return 0;
    }
    return static_cast<InternalRep *>(Tcl_GetHashValue(pEntry));
}

void
HandleNameToRepMap::erase (Tcl_HashEntry *pNameEntry)
{
    Tcl_DeleteHashEntry(pNameEntry);
}
